import logging.config

# load config from file
# logging.config.fileConfig('logging.ini', disable_existing_loggers=False)
# or, for dictConfig
config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "simple": {
            "format": "%(asctime)s %(levelname)s %(name)s: %(message)s"
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "stream": "ext://sys.stdout"
        },
        "info_file_handler": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "INFO",
            "formatter": "simple",
            "filename": "log/info.log",
            "maxBytes": 10485760,
            "backupCount": 20,
            "encoding": "utf8"
        },
        "error_file_handler": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "ERROR",
            "formatter": "simple",
            "filename": "log/error.log",
            "maxBytes": 10485760,
            "backupCount": 20,
            "encoding": "utf8"
        },
        "debug_file_handler": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "DEBUG",
            "formatter": "simple",
            "filename": "log/debug.log",
            "maxBytes": 10485760,
            "backupCount": 20,
            "encoding": "utf8"
        }
    },
    "loggers": {
        "my_module": {
            "level": "ERROR",
            "handlers": ["console"],
            "propagate": "no"
        }
    },
    "root": {
        "level": "INFO",
        "handlers": ["console", "info_file_handler", "error_file_handler", "debug_file_handler"]
    }
}



"""
Exception raised when the requested resource(route, file, database record) is no longer available and will not be
available again
"""


class GoneException(BaseException):
    """
    GoneException
    """
    pass



"""
CallbackResolver service module
"""
import os
import re
import logging
from collections import OrderedDict
from typing import Union, List, Tuple, Dict

from django.core.exceptions import RequestDataTooBig
from django.http import HttpResponse
from django.http.request import HttpRequest
from django.views.generic import TemplateView
from django.views.decorators.csrf import csrf_exempt
from raven.contrib.django.raven_compat.models import client

from rinzler.core.route_mapping import RouteMapping
from rinzler.core.response import Response
from rinzler.exceptions.conflict_exception import ConflictException
from rinzler.exceptions.content_too_large_exception import ContentTooLargeException
from rinzler.exceptions.gone_exception import GoneException
from rinzler.exceptions.internal_exception import InternalException
from rinzler.exceptions.invalid_input_exception import InvalidInputException
from rinzler.exceptions.auth_exception import AuthException
from rinzler.exceptions.not_allowed_exception import NotAllowedException
from rinzler.exceptions.not_found_exception import NotFoundException
from rinzler.exceptions.service_unavailable_exception import ServiceUnavailableException
from rinzler.exceptions.unacceptable_input_exception import UnacceptableInputException
from rinzler.exceptions.unauthorized_exception import UnauthorizedException


class Router(TemplateView):
    """
    Router
    """
    __request = None
    __callable = None
    __app = dict()
    __route = str()
    __end_point_uri = str()
    __uri = str()
    __method = str()
    __bound_routes = dict()
    __auth_service = None
    __response_callback = None
    __url_params_like = str()
    __url_params: Dict[str, str] = dict()
    __allowed_headers = "Authorization, Content-Type, If-Match, If-Modified-Since, If-None-Match, If-Unmodified-Since,"\
                        " Origin, X-GitHub-OTP, X-Requested-With, Content-Checksum"
    __allowed_methods = "GET,POST,PUT,PATCH,DELETE,HEAD,OPTIONS"

    def __init__(self, route, controller):
        super(Router, self).__init__()
        self.__route = route
        self.__callable = controller

    def flush(self):
        """
        Flushes this objects to prevent memory leaking across requests
        """
        self.__request = None
        self.__request = None
        self.__request = None
        self.__request = None
        self.__uri = None
        self.__method = None
        self.__bound_routes = dict()

    @csrf_exempt
    def route(self, request: HttpRequest):
        """
        Prepares for the CallBackResolver and handles the response and exceptions
        :param request HttpRequest
        :rtype: HttpResponse
        """
        self.flush()
        self.__request = request
        self.__uri = request.path[1:]
        self.__method = request.method
        self.__bound_routes = dict()
        self.register('log', logging.getLogger(os.urandom(3).hex().upper()))
        self.register('router', RouteMapping())

        self.__app['router'].flush_routes()
        routes = self.__callable().connect(self.__app)

        self.__bound_routes = routes['router'].get__routes()

        request_headers = request.META
        if 'HTTP_USER_AGENT' in request_headers:
            indent = 2 if re.match("[Mozilla]{7}", request_headers['HTTP_USER_AGENT']) else 0
        else:
            indent = 0

        if self.set_end_point_uri() is False:
            return self.set_response_headers(self.no_route_found(self.__request).render(indent))

        acutal_params = self.get_url_params(self.get_end_point_uri())
        response = HttpResponse(None)
        try:
            response = self.exec_route_callback(acutal_params)
        except InvalidInputException:
            client.captureException()
            self.__app['log'].error("< 400", exc_info=True)
            response = Response(None, status=400)
        except UnauthorizedException:
            client.captureException()
            self.__app['log'].error("< 401", exc_info=True)
            response = Response(None, status=401)
        except AuthException:
            client.captureException()
            self.__app['log'].error("< 403", exc_info=True)
            response = Response(None, status=403)
        except NotFoundException:
            self.__app['log'].error("< 404", exc_info=True)
            response = Response(None, status=404)
        except NotAllowedException:
            self.__app['log'].error("< 405", exc_info=True)
            response = Response(None, status=405)
        except UnacceptableInputException:
            client.captureException()
            self.__app['log'].error("< 406", exc_info=True)
            response = Response(None, status=406)
        except ConflictException:
            client.captureException()
            self.__app['log'].error("< 409", exc_info=True)
            response = Response(None, status=409)
        except GoneException:
            client.captureException()
            self.__app['log'].error("< 410", exc_info=True)
            response = Response(None, status=410)
        except RequestDataTooBig or ContentTooLargeException:
            client.captureException()
            self.__app['log'].error("< 413", exc_info=True)
            response = Response(None, status=413)
        except BaseException or InternalException:
            client.captureException()
            self.__app['log'].error("< 500", exc_info=True)
            response = Response(None, status=500)
        except ServiceUnavailableException:
            client.captureException()
            self.__app['log'].error("< 503", exc_info=True)
            response = Response(None, status=503)
        finally:
            if type(response) == Response:
                self.call_response_callback(
                    response=response, method=request.method, route=self.__route, url=self.__uri,
                    url_params_like=self.__url_params_like, url_params=self.__url_params, body=request.body,
                    app_name=routes['app_name'], auth_data=routes.get('auth_data'),
                    client_ips=self.get_client_ip(request.META)
                )
                return self.set_response_headers(response.render(indent))
            else:
                return self.set_response_headers(response)

    def exec_route_callback(self, actual_params) -> Response or object:
        """
        Executes the resolved end-point callback, or its fallback
        :param actual_params dict
        :rtype: Response or object
        """
        self.__url_params_like = str()
        self.__url_params = dict()
        if self.__method.lower() in self.__bound_routes:
            for bound in self.__bound_routes[self.__method.lower()]:

                route = list(bound)[0]
                expected_params = self.get_url_params(route)

                if self.request_matches_route(self.get_end_point_uri(), route):
                    self.__app['log'].info("> {0} {1}".format(self.__method, self.__uri))
                    authenticate = self.authenticate(route, actual_params)
                    if authenticate:
                        self.__app['log'].debug(
                            "%s(%d) %s" % ("body ", len(self.__request.body), self.__request.body.decode('utf-8'))
                        )
                        pattern_params = self.get_callback_pattern(expected_params, actual_params)
                        self.__url_params_like = route
                        self.__url_params = pattern_params
                        return bound[route](self.__request, self.__app, **pattern_params)
                    else:
                        return authenticate

        if self.__method == "OPTIONS":
            self.__app['log'].info("Route matched: {0} {1}".format(self.__method, self.__uri))
            return self.default_route_options(self.__request)

        if self.__route == '' and self.__uri == '':
            return self.welcome_page(self.__request)
        else:
            return self.no_route_found(self.__request)

    def request_matches_route(self, actual_route: str, expected_route: str):
        """
        Determines whether a route matches the actual requested route or not
        :param actual_route str
        :param expected_route
        :rtype: Boolean
        """
        expected_params = self.get_url_params(expected_route)
        actual_params = self.get_url_params(actual_route)
        i = 0

        if len(expected_params) == len(actual_params):
            for param in actual_params:
                if expected_params[i][0] != "{":
                    if param != expected_params[i]:
                        return False
                i += 1
        else:
            return False

        return True

    def authenticate(self, bound_route, actual_params) -> object:
        """
        Runs the pre-defined authenticaton method
        :param bound_route str route matched
        :param actual_params dict actual url parameters
        :rtype: object
        """
        if self.__auth_service is not None:
            auth_route = "{0}_{1}{2}".format(self.__method, self.__route, bound_route)
            auth_data = self.__auth_service.authenticate(self.__request, auth_route, actual_params)
            self.__app['auth_data'] = auth_data

        return True

    @staticmethod
    def get_callback_pattern(expected_params, actual_params):
        """
        Assembles a dictionary whith the parameters schema defined for this route
        :param expected_params dict parameters schema defined for this route
        :param actual_params dict actual url parameters
        :rtype: dict
        """
        pattern = dict()
        key = 0
        for exp_param in expected_params:
            if exp_param[0] == '{' and exp_param[-1:] == '}':
                pattern[exp_param[1:-1]] = actual_params[key]
            key = key + 1
        return pattern

    @staticmethod
    def get_url_params(end_point):
        """
        Gets route parameters as dictionary
        :param end_point str target route
        :rtype: dict
        """
        var_params = end_point.split('/')

        if len(var_params) == 1 and var_params[0] == '':
            return []

        elif len(var_params) == 1 and var_params[0] != '':
            return [var_params[0]]
        else:
            params = list()
            for param in var_params:
                if len(param) > 0:
                    params.append(param)
            return params

    def get_end_point_uri(self):
        """
        Returns the value of __end_point_uri
        :rtype: str
        """
        return self.__end_point_uri

    def set_end_point_uri(self):
        """
        Gets the route from the accessed URL
        :rtype: str
        """
        expected_parts = self.__route.split("/")
        actual_parts = self.__uri.split("/")

        i = 0
        for part in expected_parts:
            if part != actual_parts[i]:
                return False
            i = i + 1

        uri_prefix = len(self.__route)
        self.__end_point_uri = self.__uri[uri_prefix:]
        return True

    def no_route_found(self, request):
        """
        Default callback for route not found
        :param request HttpRequest
        :rtype: Response
        """
        response_obj = OrderedDict()
        response_obj["status"] = False
        response_obj["exceptions"] = {
            "message": "No route found for {0} {1}".format(self.__method, self.__uri),
        }
        response_obj["request"] = {
            "method": self.__method,
            "path_info": self.__uri,
            "content": request.body.decode("utf-8")
        }
        response_obj["message"] = "We are sorry, but something went terribly wrong."

        return Response(response_obj, content_type="application/json", status=404, charset="utf-8")

    def welcome_page(self, request):
        """
        Defaulf welcome page when the route / is note mapped yet
        :param request
        :rtype: HttpResponse
        """
        message = "HTTP/1.1 200 OK RINZLER FRAMEWORK"
        return HttpResponse(
            "<center><h1>{0}({1})</h1></center>".format(message, self.__app['app_name']),
            content_type="text/html", charset="utf-8"
        )

    @staticmethod
    def default_route_options(request: HttpRequest):
        """
        Default callback for OPTIONS request
        :param request HttpRequest
        :rtype: Response
        """
        response_obj = OrderedDict()

        response_obj["status"] = True
        response_obj["data"] = "Ok"

        return Response(response_obj, content_type="application/json", charset="utf-8")

    def set_response_headers(self, response: HttpResponse) -> HttpResponse:
        """
        Appends default headers to every response returned by the API
        :param response HttpResponse
        :rtype: HttpResponse
        """
        public_name = os.environ.get('SERVER_PUBLIC_NAME')
        response_headers = {
            'access-control-allow-headers': self.__allowed_headers,
            'access-control-allow-methods': self.__allowed_methods,
            'access-control-allow-origin': "*",
            'access-control-allow-credentials': True,
            'www-authenticate': "Bearer",
            'server-public-name': public_name if public_name is not None else "No one",
            'user-info': "Rinzler Framework rulez!"
        }

        for key in response_headers:
            response[key] = response_headers[key]

        status = response.status_code
        if status != 404:
            self.__app['log'].info("< {0}".format(status))

        return response

    def auth_config(self, auth_service: object):
        """
        Sets the authenticator service.
        :param auth_service object
        :rtype: Router
        """
        self.__auth_service = auth_service
        return self

    def register(self, name: str, handler: object, force=False):
        """
        Registers a sevico to the APP object
        :param name: str
        :param handler: object
        :param force: bool
        :return: self
        """
        if name in self.__app:
            if force is True:
                self.__app[name] = handler
        else:
            self.__app[name] = handler

        return self

    def response_callback_config(self, callback: object):
        """
        Sets the authenticator service.
        :param callback object
        :rtype: Router
        """
        self.__response_callback = callback
        return self

    def call_response_callback(self, **kwargs) -> bool:
        """
        Calls the response callback configured to this instance
        :param kwargs: the parameters you'd like to pass
        :rtype: bool
        """
        if self.__response_callback:
            self.__response_callback.call(**kwargs)

        return True

    @staticmethod
    def get_client_ip(meta: dict) -> Union[Tuple[List[str], str, str], None]:
        """
        Returns the client ip address from the request headers
        Return the headers: HTTP_X_FORWARDED_FOR, HTTP_X_REAL_IP, REMOTE_ADDR
        """
        try:
            ip_f = meta.get("HTTP_X_FORWARDED_FOR", None)
            ip_list = list()
            if ip_f:
                ip_list: List[str] = [i.strip() for i in ip_f.split(",")]

            return ip_list, meta.get("HTTP_X_REAL_IP"), meta.get("REMOTE_ADDR")
        except (KeyError, IndexError):
            return None



"""
Exception raised when a body request can not be decode from JSON and/or it's content doesn't match the minimum
specified values by the app
"""


class InvalidInputException(BaseException):
    """
    InvalidInputException
    """
    pass



import logging

from .logger.log import setup_logging
from .core.url_assembler import UrlAssembler

__name__ = "Rinzler REST Framework"
__version__ = "1.24.3"
__author__ = "Rinzler"


def boot(app_name):
    """
    Start Rinzler App
    :param app_name: str
    :return: dict
    """
    setup_logging()
    url_assembler = UrlAssembler().set_app_name(app_name)

    logger = logging.getLogger(app_name)

    logger.info("App booted =)")

    return {
        'url_assembler': url_assembler,
        'app_name': app_name,
        'log': logger
    }



"""
Exception raised when the theres a conflict in the request, such as an edit conflict between multiple simultaneous
updates
"""


class ConflictException(BaseException):
    """
    ConflictException
    """
    pass



"""
Interface for HttpResponse
"""
import json
from collections import OrderedDict
from django.http.response import HttpResponse


class Response(object):
    """
    Response
    """
    __content = OrderedDict
    __content_type = str()
    __charset = str()
    __kwargs = dict
    __indent = 0

    def __init__(self, content, content_type="application/json", charset="utf-8", **kwargs):
        """
        Constructor
        :param content str conteúdo da resposta
        :param content_type str content-type da resposta
        :param kwargs dict
        :rtype: None
        """
        self.__content = content
        self.__content_type = content_type
        self.__charset = charset
        self.__kwargs = kwargs

    def render(self, indent=0):
        """
        Renders a HttpResponse for the ongoing request
        :param indent int
        :rtype: HttpResponse
        """
        self.__indent = indent
        return HttpResponse(
            self.__str__(), content_type=self.__content_type, charset=self.__charset, **self.__kwargs
        )

    def __str__(self):
        if self.__indent > 0:
            if self.__content is not None:
                return json.dumps(self.__content, indent=self.__indent, sort_keys=False)
            else:
                return str()
        else:
            if self.__content is not None:
                return json.dumps(self.__content, sort_keys=False)
            else:
                return str()

    def __repr__(self):
        return self.__str__()

    def get_decoded(self):
        """
        Returns a decoded instance of this object
        :rtype: object
        """
        return json.loads(self.__content)



from django.conf.urls import url

from rinzler.core.router import Router


class UrlAssembler(object):
    """
    UrlAssembler
    """
    auth_service = None
    base_path = str()
    logger = object
    app_name = str()
    response_callback = object

    def mount(self, route, callback):
        """
        Mounts a route with the given params
        :rtype: urls
        """
        return url(
            r'{0}'.format(route),
            Router(route, callback)
                .register('app_name', self.app_name)
                .auth_config(self.auth_service)
                .response_callback_config(self.response_callback)
                .route
        )

    def set_auth_service(self, auth_service):
        """
        Sets the authentication service
        :rtype: object
        """
        self.auth_service = auth_service
        return self

    def set_app_name(self, app_name: str):
        """
        Sets the current running app's name
        :param app_name: str
        :return: self
        """
        self.app_name = app_name
        return self

    def set_response_callback(self, callback):
        """
        Sets a callback to be called after every Response
        """
        self.response_callback = callback
        return self



"""
Exception raised when the App tries to comunicate with third-party online services and this comunications fails
"""


class ServiceUnavailableException(BaseException):
    """
    ServiceUnavailableException
    """
    pass



"""
Route to callback mapper
"""


class RouteMapping(object):
    """
    RouteMapping
    """
    __routes = dict()

    def __init__(self):
        self.__routes = dict()

    def get(self, route: str, callback: object()):
        """
        Binds a GET route with the given callback 
        :rtype: object
        """
        self.__set_route('get', {route: callback})
        return RouteMapping

    def post(self, route: str, callback: object()):
        """
        Binds a POST route with the given callback
        :rtype: object
        """
        self.__set_route('post', {route: callback})
        return RouteMapping

    def put(self, route: str, callback: object()):
        """
        Binds a PUT route with the given callback
        :rtype: object
        """
        self.__set_route('put', {route: callback})
        return RouteMapping

    def patch(self, route: str, callback: object()):
        """
        Binds a PATCH route with the given callback
        :rtype: object
        """
        self.__set_route('patch', {route: callback})
        return RouteMapping

    def delete(self, route: str, callback: object()):
        """
        Binds a PUT route with the given callback
        :rtype: object
        """
        self.__set_route('delete', {route: callback})
        return RouteMapping

    def head(self, route: str, callback: object()):
        """
        Binds a HEAD route with the given callback
        :rtype: object
        """
        self.__set_route('head', {route: callback})
        return RouteMapping

    def options(self, route: str, callback: object()):
        """
        Binds a OPTIONS route with the given callback
        :rtype: object
        """
        self.__set_route('options', {route: callback})
        return RouteMapping

    def __set_route(self, type_route, route):
        """
        Sets the given type_route and route to the route mapping
        :rtype: object
        """
        if type_route in self.__routes:
            if not self.verify_route_already_bound(type_route, route):
                self.__routes[type_route].append(route)
        else:
            self.__routes[type_route] = [route]
        return RouteMapping

    def verify_route_already_bound(self, type_route: str, route: dict) -> bool:
        """

        :param type_route: str
        :param route: dict
        :return: bool
        """
        for bound_route in self.__routes[type_route]:
            bound_key = list(bound_route.keys())[0]
            route_key = list(route.keys())[0]
            if bound_key == route_key:
                return True

    def get__routes(self):
        """
        Gets the mapped routes
        :rtype: dict
        """
        return self.__routes

    def flush_routes(self):
        """

        :return: self\
        """
        self.__routes = dict()
        return self



"""
Exception raised when a body request could be parsed normally, but it's contents doesn't satisfies the specified
values by the app
"""


class UnacceptableInputException(BaseException):
    """
    UnacceptableInputException
    """
    pass



import os
import logging.config

import yaml

from rinzler.logger import config


def setup_logging(default_path='logging.yaml', env_key='LOG_CFG'):
    """
    Setup logging configuration
    """
    path = default_path
    value = os.getenv(env_key, None)
    if value:
        path = value
    if os.path.exists(path):
        with open(path, 'rt') as f:
            configs = yaml.safe_load(f.read())
        logging.config.dictConfig(configs)
    else:
        logging.config.dictConfig(config)



"""
Exception raised when the app fails to process the request and due to an error it couldn't handle correctly
"""


class InternalException(BaseException):
    """
    InternalException
    """
    pass
